@@ -15,11 +15,11 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include "sys-unistd.h" /* <unistd.h> */
 
 #include <stdlib.h>
 #include <string.h>
 #include <fcntl.h>
-#include <unistd.h>
 
 #ifdef HAVE_SYSLOG_H
 # include <syslog.h>
@@ -164,8 +164,15 @@ typedef struct {
     format_fields *default_format;/* allocated if default format */
 } plugin_data;
 
+typedef void(esc_fn_t)(buffer * restrict b, const char * restrict s, size_t len);
+
+typedef enum {
+    BS_ESCAPE_DEFAULT
+   ,BS_ESCAPE_JSON
+} buffer_bs_escape_t;
+
 INIT_FUNC(mod_accesslog_init) {
-    return calloc(1, sizeof(plugin_data));
+    return ck_calloc(1, sizeof(plugin_data));
 }
 
 __attribute_cold__
@@ -261,8 +268,7 @@ static format_fields * accesslog_parse_format(const char * const format, const u
     } while (++i < flen);
 
     format_fields * const fields =
-      malloc(sizeof(format_fields) + ((used+1) * sizeof(format_field)));
-    force_assert(fields);
+      ck_malloc(sizeof(format_fields) + ((used+1) * sizeof(format_field)));
     memset(fields, 0, sizeof(format_fields));
     memcpy(fields->ptr, fptr, (used+1) * sizeof(format_field));
     return fields;
@@ -372,6 +378,7 @@ SETDEFAULTS_FUNC(mod_accesslog_set_defaults) {
 
     /* process and validate config directives
      * (init i to 0 if global context; to 1 to skip empty global context) */
+    int uses_syslog = 0;
     for (int i = !p->cvlist[0].v.u2[1]; i < p->nconfig; ++i) {
         config_plugin_value_t *cpv = p->cvlist + p->cvlist[i].v.u2[0];
         int use_syslog = 0;
@@ -434,6 +441,7 @@ SETDEFAULTS_FUNC(mod_accesslog_set_defaults) {
 
         if (srv->srvconf.preflight_check) continue;
 
+        uses_syslog |= use_syslog;
         if (use_syslog) continue; /* ignore the next checks */
         cpv = cpvfile; /* accesslog.filename handled after preflight_check */
         if (NULL == cpv) continue;
@@ -447,7 +455,11 @@ SETDEFAULTS_FUNC(mod_accesslog_set_defaults) {
         }
     }
 
+  #ifdef HAVE_SYSLOG_H
     p->defaults.syslog_level = LOG_INFO;
+    if (uses_syslog)
+        fdlog_openlog(srv->errh, srv->srvconf.syslog_facility);
+  #endif
 
     /* initialize p->defaults from global config context */
     if (p->nconfig > 0 && p->cvlist->v.u2[1]) {
@@ -563,14 +575,29 @@ static format_fields * mod_accesslog_process_format(const char * const format, c
 					}
 				} else if (FORMAT_HEADER == f->field
 				           || FORMAT_RESPONSE_HEADER == f->field) {
-					f->opt = http_header_hkey_get(BUF_PTR_LEN(fstr));
-				} else if (FORMAT_REMOTE_HOST == f->field) {
+					if (buffer_is_blank(fstr)) f->field = FORMAT_LITERAL; /*(blank)*/
+					else f->opt = http_header_hkey_get(BUF_PTR_LEN(fstr));
+				} else if (FORMAT_REMOTE_HOST == f->field
+				           || FORMAT_REMOTE_ADDR == f->field) {
 					f->field = FORMAT_REMOTE_ADDR;
+					const char * const ptr = fstr->ptr;
+					if (buffer_is_blank(fstr)) {
+					}
+					else if (0 == strcmp(ptr, "mask"))
+						f->opt = 1; /* mask IP addr lower bits (partial anon) */
+					else {
+						log_error(srv->errh, __FILE__, __LINE__,
+							"invalid format %%{mask}a: %s", format);
+						mod_accesslog_free_format_fields(parsed_format);
+						return NULL;
+					}
 				} else if (FORMAT_REMOTE_USER == f->field) {
 					f->field = FORMAT_ENV;
 					buffer_copy_string_len(fstr, CONST_STR_LEN("REMOTE_USER"));
-				} else if (FORMAT_NOTE == f->field) {
-					f->field = FORMAT_ENV;
+				} else if (FORMAT_ENV == f->field
+				           || FORMAT_NOTE == f->field) {
+					if (buffer_is_blank(fstr)) f->field = FORMAT_LITERAL; /*(blank)*/
+					else f->field = FORMAT_ENV;
 				}
 			}
 
@@ -584,14 +611,12 @@ TRIGGER_FUNC(log_access_periodic_flush) {
     return HANDLER_GO_ON;
 }
 
-#define accesslog_append_escaped buffer_append_bs_escaped
-
 static void
 accesslog_append_buffer (buffer * const restrict dest,
-                         const buffer * const restrict b, const int esc)
+                         const buffer * const restrict b, esc_fn_t esc_fn)
 {
     if (!buffer_string_is_empty(b))
-        accesslog_append_escaped(dest, BUF_PTR_LEN(b), esc);
+        esc_fn(dest, BUF_PTR_LEN(b));
     else
         buffer_append_char(dest, '-');
 }
@@ -610,7 +635,8 @@ __attribute_noinline__
 static void
 accesslog_append_cookie (buffer * const restrict dest,
                          const request_st * const restrict r,
-                         const buffer * const restrict name, const int esc)
+                         const buffer * const restrict name,
+                         esc_fn_t esc_fn)
 {
     const buffer * const vb =
       http_header_request_get(r, HTTP_HEADER_COOKIE, CONST_STR_LEN("Cookie"));
@@ -625,7 +651,7 @@ accesslog_append_cookie (buffer * const restrict dest,
             for (str = v; *str != '\0' && *str != ';'; ++str) ;
             if (str == v) break;
             do { --str; } while (str > v && (*str == ' ' || *str == '\t'));
-            accesslog_append_escaped(dest, v, str - v + 1, esc);
+            esc_fn(dest, v, str - v + 1);
             break;
         }
         else {
@@ -719,11 +745,19 @@ accesslog_append_time (buffer * const b, const request_st * const r,
 					buffer_clear(ts_accesslog_str);
 				      #if defined(HAVE_STRUCT_TM_GMTOFF)
 					buffer_append_strftime(ts_accesslog_str,
+					                     #ifdef __MINGW32__
+					                       fmt ? fmt : "[%d/%b/%Y:%H:%M:%S %z]",
+					                     #else
 					                       fmt ? fmt : "[%d/%b/%Y:%T %z]",
+					                     #endif
 					                       localtime64_r(&t, &tm));
 				      #else /* HAVE_STRUCT_TM_GMTOFF */
 					buffer_append_strftime(ts_accesslog_str,
+					                     #ifdef ___MINGW32__
+					                       fmt ? fmt : "[%d/%b/%Y:%H:%M:%S +0000]",
+					                     #else
 					                       fmt ? fmt : "[%d/%b/%Y:%T +0000]",
+					                     #endif
 					                       gmtime64_r(&t, &tm));
 				      #endif /* HAVE_STRUCT_TM_GMTOFF */
 					buffer_append_string_buffer(b, ts_accesslog_str);
@@ -754,17 +788,59 @@ accesslog_append_time (buffer * const b, const request_st * const r,
 			return flush;
 }
 
+__attribute_noinline__
+static void
+accesslog_append_remote_addr_masked (buffer * const b, const request_st * const r)
+{
+    /* mask lower bits of IP address string for logging
+     * (similar to masking policy applied by Google Analytics
+     *  https://support.google.com/analytics/answer/2763052) */
+    /* r->dst_addr_buf is normalized and valid string; operate on known valid */
+    const char * const s = r->dst_addr_buf->ptr;
+    uint32_t i = 0;
+    switch (sock_addr_get_family(r->dst_addr)) {
+     #ifdef HAVE_IPV6
+      case AF_INET6:
+        if (__builtin_expect( (s[0] != ':'), 1)
+            || !IN6_IS_ADDR_V4MAPPED(&((sock_addr*)r->dst_addr)->ipv6.sin6_addr)
+            || NULL == strchr(s, '.')) {
+            /* IPv6: mask final 10 octets (80 bits) of address; keep 6 octets */
+            /* Note: treat string starting w/ "::..." as "::" even if "::x:..."
+             * (rather than special-casing; does not detect "::x:..." besides
+             *  v4mapped "::ffff:1.2.3.4" condition check above) */
+            for (int j=0; s[i] != ':' || ((j+=2) != 6 && s[i+1] != ':'); ++i) ;
+            buffer_append_str2(b, s, i+1, CONST_STR_LEN(":"));
+            break;
+        }
+        /* IN6_IS_ADDR_V4MAPPED() */
+        __attribute_fallthrough__
+     #endif
+      case AF_INET:
+        /* IPv4: mask final octet (8 bits) of address */
+       #ifdef __COVERITY__
+        force_assert(buffer_clen(r->dst_addr_buf) > 2);
+        force_assert(strchr(s, '.') != NULL);
+       #endif
+        for (i = buffer_clen(r->dst_addr_buf)-1; s[--i] != '.'; ) ;
+        buffer_append_str2(b, s, i+1, CONST_STR_LEN("0"));
+        break;
+      default:
+        buffer_append_buffer(b, r->dst_addr_buf);
+        break;
+    }
+}
+
 __attribute_cold__
 __attribute_noinline__
 static void
 log_access_record_cold (buffer * const b, const request_st * const r,
-                        const format_field * const f, const int esc)
+                        const format_field * const f, esc_fn_t esc_fn)
 {
     connection * const con = r->con;
     switch (f->field) {
       case FORMAT_SERVER_PORT:
         if (f->opt & FORMAT_FLAG_PORT_REMOTE) {
-            buffer_append_int(b, sock_addr_get_port(&con->dst_addr));
+            buffer_append_int(b, sock_addr_get_port(r->dst_addr));
             break;
         }
         /* else if (f->opt & FORMAT_FLAG_PORT_LOCAL) *//*(default)*/
@@ -797,16 +873,15 @@ log_access_record_cold (buffer * const b, const request_st * const r,
         {
             const uint32_t len = buffer_clen(&r->target);
             const char * const qmark = memchr(r->target.ptr, '?', len);
-            accesslog_append_escaped(b, r->target.ptr,
-                                     qmark ? (uint32_t)(qmark - r->target.ptr)
-                                           : len, esc);
+            esc_fn(b, r->target.ptr,
+                   qmark ? (uint32_t)(qmark - r->target.ptr) : len);
         }
         break;
       case FORMAT_QUERY_STRING:
-        accesslog_append_escaped(b, BUF_PTR_LEN(&r->uri.query), esc);
+        esc_fn(b, BUF_PTR_LEN(&r->uri.query));
         break;
       case FORMAT_FILENAME:
-        accesslog_append_buffer(b, &r->physical.path, esc);
+        accesslog_append_buffer(b, &r->physical.path, esc_fn);
         break;
       case FORMAT_CONNECTION_STATUS:
         buffer_append_char(b, (r->state == CON_STATE_RESPONSE_END)
@@ -829,7 +904,7 @@ log_access_record_cold (buffer * const b, const request_st * const r,
     }
 }
 
-static int log_access_record (const request_st * const r, buffer * const b, format_fields * const parsed_format, const buffer_bs_escape_t esc) {
+static int log_access_record (const request_st * const r, buffer * const b, format_fields * const parsed_format, esc_fn_t esc) {
 	const buffer *vb;
 	unix_timespec64_t ts = { 0, 0 };
 	int flush = 0;
@@ -871,7 +946,10 @@ static int log_access_record (const request_st * const r, buffer * const b, form
 			/*case FORMAT_REMOTE_HOST:*/
 		  #endif
 			case FORMAT_REMOTE_ADDR:
-				buffer_append_string_buffer(b, &r->con->dst_addr_buf);
+				if (!f->opt)
+					buffer_append_string_buffer(b, r->dst_addr_buf);
+				else
+					accesslog_append_remote_addr_masked(b, r);
 				break;
 			case FORMAT_HTTP_HOST:
 				accesslog_append_buffer(b, &r->uri.authority, esc);
@@ -880,7 +958,7 @@ static int log_access_record (const request_st * const r, buffer * const b, form
 				/*(attempt to reconstruct request line)*/
 				http_method_append(b, r->http_method);
 				buffer_append_char(b, ' ');
-				accesslog_append_escaped(b, BUF_PTR_LEN(&r->target_orig), esc);
+				esc(b, BUF_PTR_LEN(&r->target_orig));
 				buffer_append_char(b, ' ');
 				http_version_append(b, r->http_version);
 				break;
@@ -930,8 +1008,11 @@ REQUESTDONE_FUNC(log_access_write) {
       ? (buffer_clear(r->tmp_buf), r->tmp_buf)
       : &fdlog->b;
 
+    esc_fn_t * const esc_fn = !p->conf.escaping
+      ? buffer_append_bs_escaped
+      : buffer_append_bs_escaped_json;
     const int flush =
-      log_access_record(r, b, p->conf.parsed_format, p->conf.escaping);
+      log_access_record(r, b, p->conf.parsed_format, esc_fn);
 
   #ifdef HAVE_SYSLOG_H
     if (p->conf.use_syslog) {
@@ -955,6 +1036,8 @@ REQUESTDONE_FUNC(log_access_write) {
 }
 
 
+__attribute_cold__
+__declspec_dllexport__
 int mod_accesslog_plugin_init(plugin *p);
 int mod_accesslog_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
